// screensaver.cpp : Defines the class behaviors for the application.
//

#include "stdafx.h"
#include "screensaver.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CScreensaverApp

BEGIN_MESSAGE_MAP(CScreensaverApp, CWinApp)
	//{{AFX_MSG_MAP(CScreensaverApp)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CScreensaverApp construction

CScreensaverApp::CScreensaverApp()
{
	// Place all significant initialization in InitInstance
}

// virtual
CScreensaverApp::~CScreensaverApp()
{
	// Intentionally empty
}

/////////////////////////////////////////////////////////////////////////////
// CScreensaverApp initialization
BOOL CScreensaverApp::InitInstance()
{
	// Standard initialization
	// If you are not using these features and wish to reduce the size
	//  of your final executable, you should remove from the following
	//  the specific initialization routines you do not need.

#ifdef _AFXDLL
	Enable3dControls();			// Call this when using MFC in a shared DLL
#else
	Enable3dControlsStatic();	// Call this when linking to MFC statically
#endif

	// Change the registry key under which our settings are stored.
	SetRegistryKey( applicationRegistryKey() );

	// When we are invoked, there are a number of command line parameters
	// which govern how we are to operate.  They are as follows:
	//
	//		/s			- Run the screensaver in fullscreen mode
	//		/c			- Run any configuration settings
	//		/a			- Change the password
	//		/p <HWND>	- Run the screensaver in preview mode with HWND
	//					  being the parent window we are to run within.
	//
	// As a special debugging aid, we can test the preview option
	// by running the screensaver in a debugger using /testpreview.
	// In this case, the app will create a parent window internally
	// solely to test the preview.  To test fullscreen, simply supply
	// the /s switch as normal.

	// Determine the command line parameters
	CString		cmdLine=m_lpCmdLine;
	cmdLine.MakeLower();
	if ( cmdLine.Find(_T("/s")) != -1 )
	{
		// Run screensaver proper
		runScreensaver();
	}
	else if ( cmdLine.Find(_T("/c")) != -1 )
	{
		// Run configuration
		runConfiguration();
	}
	else if ( cmdLine.Find(_T("/a")) != -1 )
	{
		// Run password change
		runChangePassword();
	}
	else if ( cmdLine.Find(_T("/testpreview")) != -1 )
	{
		// Run the preview but in test mode.
		// For this, we need to create a
		// window to act as the parent of
		// the screensaver.
		HWND	hParentWnd=createHelperWindow();
		ASSERT( ::IsWindow(hParentWnd) );
		runPreview( hParentWnd );
		::DestroyWindow( hParentWnd );
	}
	else if ( cmdLine.Find(_T("/p")) != -1 )
	{
		// Run preview but we need to extract the HWND
		// of the parent window
		const int	index=cmdLine.Find( _T("/p") );
		ASSERT( index!=-1 );

		// We don't bother checking that the parent window is
		// invalid - the preview method will verify and early
		// out if there's a problem.  For debugging, we just
		// do a sanity check
		HWND		hParentWnd=(HWND)atoi( cmdLine.Mid(index+3) );
		ASSERT( ::IsWindow(hParentWnd) );
		runPreview( hParentWnd );
	}
	else
	{
		TRACE( "CScreensaverApp::InitInstance() Unrecognised command line parameters: '%s'\n",(LPCTSTR)cmdLine );
		ASSERT( false );
	}

	return FALSE;
}

void CScreensaverApp::runScreensaver()
{
	// This will run the screensaver in fullscreen mode.
	// We simply handle the drudgery here and defer the
	// actual work to the concrete class.
	MSG			msg;
	int			originalSPISetting;
	bool		isPowerOn=true;

	// Allow the client to initialise itself
	HWND		hWnd=doCreateFullscreen();
	if ( hWnd!=NULL )
	{
		// We should be initialised okay so let Windows
		// know that a screensaver is running.
		::SystemParametersInfo( SPI_SCREENSAVERRUNNING,true,&originalSPISetting,0 );

		// Process whilst we have a valid window
		while ( ::IsWindow(hWnd) || !verifyPassword(hWnd) )
		{
			// Only render if we have still got power
			if ( isPowerOn )
				doRenderFullscreen();

			// Operate a standard message pump
			while ( ::PeekMessage(&msg,NULL,0,0,PM_REMOVE) )
			{
				// For Win98 and Win2K systems, we could have the
				// monitor powered off.  If this happens, Direct3D/OpenGL
				// can loose their surfaces, etc.
				if ( _winver >= 0x0400 )
				{
					if ( msg.message==WM_SYSCOMMAND && msg.wParam==SC_MONITORPOWER )
					{
						// We've lost power
						isPowerOn=false;
					}
					else if ( msg.message==WM_PAINT )
					{
						// If we get this and power was off, let the
						// client know that surfaces need to be
						// restored
						if ( !isPowerOn )
						{
							doRestoreFullscreen();
							isPowerOn=true;
						}
					}
				}

				// The normal translate and despatch stuff
				PreTranslateMessage( &msg );
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
		}

		// Ensure the window is destroyed
		if ( ::IsWindow(hWnd) )
		{
			::SendMessage( hWnd,WM_CLOSE,0,0 );
		}
		ASSERT( !::IsWindow(hWnd) );

		// Prepare for the end.
		// Restore Windows' view of what was running
		::SystemParametersInfo( SPI_SCREENSAVERRUNNING,false,&originalSPISetting,0 );

		// Allow the client to perform an orderly shutdown
		doDestroyFullscreen();
	}
}

void CScreensaverApp::runConfiguration()
{
	// Nothing really clever here - simply delegate to client
	doRunConfiguration();
}

void CScreensaverApp::runPreview( HWND hParentWnd )
{
	// This will run the preview mode.
	// As in the case of fullscreen, we handle alot
	// of the mudane stuff here and delegate the
	// juicy bits to the client.
	MSG			msg;
	bool		isPowerOn=true;

	// Verify the parent window
	if ( !::IsWindow(hParentWnd) )
	{
		TRACE( "CScreensaverApp::runPreview() Parent window is invalid!\n" );
		return;
	}

	// Wait for the parent window to become visible
	while ( ::IsWindow(hParentWnd) && !::IsWindowVisible(hParentWnd) )
	{
		// Pump away
		while ( ::PeekMessage(&msg,hParentWnd,0,0,PM_REMOVE) )
		{
			PreTranslateMessage( &msg );
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
	}

	// Allow the client to initialise itself
	HWND		hWnd=doCreatePreview( hParentWnd );
	if ( hWnd!=NULL )
	{
		// Process whilst we have a valid window
		while ( ::IsWindow(hParentWnd) && ::IsWindowVisible(hParentWnd) && ::IsWindow(hWnd) )
		{
			// Only render if we have still got power
			if ( isPowerOn )
				doRenderPreview();

			// Operate a standard message pump
			while ( ::PeekMessage(&msg,NULL,0,0,PM_REMOVE) )
			{
				// For Win98 and Win2K systems, we could have the
				// monitor powered off.  If this happens, Direct3D/OpenGL
				// can loose their surfaces, etc.
				if ( _winver >= 0x0400 )
				{
					if ( msg.message==WM_SYSCOMMAND && msg.wParam==SC_MONITORPOWER )
					{
						// We've lost power
						isPowerOn=false;
					}
					else if ( msg.message==WM_PAINT )
					{
						// If we get this and power was off, let the
						// client know that surfaces need to be
						// restored
						if ( !isPowerOn )
						{
							doRestorePreview();
							isPowerOn=true;
						}
					}
				}

				// The normal translate and despatch stuff
				PreTranslateMessage( &msg );
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
		}

		// Ensure the window is destroyed
		if ( ::IsWindow(hWnd) )
		{
			::SendMessage( hWnd,WM_CLOSE,0,0 );
		}
		ASSERT( !::IsWindow(hWnd) );

		// Allow the client to perform an orderly shutdown
		doDestroyPreview();
	}
}

void CScreensaverApp::runChangePassword()
{
	// We need to dynamically link to the password DLL since
	// it contains the function to change the screensaver
	// password.
	HINSTANCE	hInst=::LoadLibrary( _T("MPR.DLL") );
	if ( hInst==NULL )
	{
		TRACE( "CScreensaverApp::runChangePassword() Failed to load MPR.DLL\n" );
	}
	else
	{
		// Get hold of the correct function
		typedef VOID (WINAPI *PSWCHANGER)( LPCSTR lpRegKey,HWND hWnd,UINT nReserved1,UINT nReserved2 );
		PSWCHANGER	pswChanger=(PSWCHANGER)::GetProcAddress( hInst,_T("PwdChangePasswordA") );
		if ( pswChanger==NULL )
		{
			TRACE( "CScreensaverApp::runChangePassword() Failed to get address of PwdChangePasswordA\n" );
		}
		else
		{
			// Change the password
			pswChanger( _T("SCRSAVE"),::GetDesktopWindow(),0,0 );
		}

		// Free the library
		::FreeLibrary( hInst );
	}
}

bool CScreensaverApp::verifyPassword( HWND hParentWnd )
{
	// This method will optionally bring up the password
	// verification dialog and verify the password
	bool	result=false;

	// If we're running on WinNT, it will handle the password
	// verification for us so we don't need to bother with
	// anything
	OSVERSIONINFO		osv; 
	ZeroMemory( &osv,sizeof(osv) );
	osv.dwOSVersionInfoSize=sizeof(osv); 
	::GetVersionEx( &osv );
	if ( osv.dwPlatformId==VER_PLATFORM_WIN32_NT )
	{
		result=true;
	}
	else
	{
		// Use the password control panel applet
		HINSTANCE	hInst=::LoadLibrary( _T("PASSWORD.CPL") );
		if ( hInst==NULL )
		{
			TRACE( "CScreensaver::verifyPassword() Failed to load PASSWORD.CPL\n" );
		}
		else
		{
			// Get hold of the password verification function
			typedef BOOL (WINAPI *PSWVERIFY)( HWND hWnd );
			PSWVERIFY		pswVerify=(PSWVERIFY)::GetProcAddress( hInst,_T("VerifyScreenSavePwd") );
			if ( pswVerify==NULL )
			{
				TRACE( "CScreensaver::verifyPassword() Failed to get address of VerifyScreenSavePwd\n" );
			}
			else
			{
				::ShowCursor( TRUE );

				result = (pswVerify(hParentWnd)==TRUE);

				::ShowCursor( FALSE );
			}

			// Release the library
			::FreeLibrary( hInst );
		}
	}

	return result;
}

// Test preview window procedure.  Nothing special, really.
LRESULT CALLBACK PreviewHelperWndProc( HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam )
{
	switch( msg ) 
	{
		case WM_PAINT:
		{
			// Just draw a black background
			PAINTSTRUCT		ps;
			HDC				hDC=::BeginPaint( hWnd,&ps );
			::FillRect( hDC,&ps.rcPaint,(HBRUSH)::GetStockObject(BLACK_BRUSH) );
			::EndPaint( hWnd,&ps );
			break;
		}
		default:
			return ::DefWindowProc( hWnd,msg,wParam,lParam );
   }

   return 0;
}

HWND CScreensaverApp::createHelperWindow()
{
	// We need to create a little window
	// to serve as the parent to a preview
	// test.
	static bool		hasBeenRegistered=false;
	static LPCTSTR	lpszClassName=_T("ScreensaverPreviewHelperClass");
	if ( !hasBeenRegistered )
	{
		WNDCLASSEX		wc;
		
		ZeroMemory( &wc,sizeof(wc) );
		wc.cbSize			= sizeof(wc);
		wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
		wc.lpfnWndProc		= (WNDPROC) PreviewHelperWndProc;
		wc.cbClsExtra		= 0;
		wc.cbWndExtra		= 0;
		wc.hInstance		= ::AfxGetInstanceHandle();
		wc.hIcon			= NULL;
		wc.hCursor			= ::LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground	= NULL;
		wc.lpszMenuName		= NULL;
		wc.lpszClassName	= lpszClassName;

		ATOM	atom=::RegisterClassEx( &wc );
		ASSERT( atom!=0 );

		hasBeenRegistered=true;
	}

	// Choose a size that befits the current resolution
	const int	xScreen=::GetSystemMetrics( SM_CXSCREEN );
	const int	yScreen=::GetSystemMetrics( SM_CYSCREEN );

	// Set the width and height to a quarter that of the screen
	const int	w=xScreen / 4;
	const int	h=yScreen / 4;

	// Centre the window
	const int	x=(xScreen - w) / 2;
	const int	y=(yScreen - h) / 2;

	// Create the window.
	LPCTSTR		lpszTitle=_T("Screensaver Preview Test");
	DWORD		dwStyles=WS_CAPTION;
	HWND		result=::CreateWindow( lpszClassName,lpszTitle,dwStyles,x,y,w,h,NULL,NULL,::AfxGetInstanceHandle(),NULL );
	ASSERT( ::IsWindow(result) );
	::ShowWindow( result,SW_SHOW );

	return result;
}